/************************************************************************************
MODULE     :   LocBusSlave        AUTHOR    :   Tullio Grassi

DESCRIPTION :
The Local Bus interfaces different components on a board. This bus includes the data
lines, the address lines, WRITE_b (active low) and several CSn_b (active low
ChipSelect: CS1_b, CS2_b, for all the  slave devices).
The protocol foresees that when the master drives CS_b low for a particular slave,
the slave must decode the address on the Loc_Address lines, and read or write the 
Loc_Data bus depending of the WRITE_b direction.
On a write cycle, the master set WRITE_b = 0, set the address and data on the Local Bus,
and the slave will latch the data when CS_b = 0. 
The master keeps the data valid for the whole duration of the CS_b pulse. 
On a read cycle, the master set WRITE_b = 1, set the address on the Local Bus and
latch the data on the rising edge of CS_b, as the slave must prepare the data 
during the CS_b pulse. The CS_b pulse width depends on the slower slave.
A value of 300 ns is usually fine. When a slave is not selected, its data port is on Hi-Z.
The design is fully synchronous with clk. 
LocalBus lines are syncronized (this adds a few ticks of delay for edge detection).

!!!! When adding registers, remember that some of them are read-only,
some are write-only. Do not assign addresses already used !!!!!
**********************************************************************************/

`include "define.v"
module LocBusSlave  (      
// these ports connect to the LocalBus
  input  LocCS_b,  LocWRITE_b,       
  input  [`LB_Addr_Size-1:0]  LocAddr,  // The Address bus.
  inout  [`REG_WIDTH-1:0]     LocData,  // The bidir Data bus 
// these ports connect to the rest of the FPGA
  input clk,  rst, RxDLL_lock,  TTC_DLL_lock,  run_status,  BunchEr, 
  input HistogramsDone,
  input [11:0] latch_bucket, count_bucket,
  input [2:0] UnlockCnt,
  input [15:0] DLL_status,
               RxClkCnt, RxBC0Cnt, TClkCnt, TxBC0Cnt,
  input [8:1]  TPSpyFifoEmpty,  TPSpyFifoFull, 
  input [35:0]  TPSpyFIFO1_out,  TPSpyFIFO2_out, TPSpyFIFO3_out,
  input [7:0]  LastTTCcommand,
  input [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
               TTC_test4, TTC_test5, TTC_test6, TTC_test7,
               TTC_test8, TTC_test9, TTC_testA, TTC_testB,
               TTC_testC, TTC_testD, TTC_testE, TTC_testF,
  input [15:0] hist0_value,hist1_value,hist2_value,hist3_value,histsum_value,
  input [12:0] sum0_count, sum1_count, sum2_count, sum3_count,

  output reg [47:0] DestMAC,
  output reg [47:0] HLXMAC,
  output reg read_count,
  output reg hist_localbus_read, HZero,
  output reg [11:0] hist_localbus_bin,
  output reg [4:0] hist_localbus_select,
  output reg LEDdoBC0,
  output reg [7:0] TTC_BC0_pattern,
  output  GateTTCcommands, 
  output reg VME_Hard_rst, VME_Soft_rst,
             VME_Start, VME_Stop, VME_LVDSout,
  output reg [7:0] mux_out,  
  output reg [8:1] TPSpyFifoRd,
  output reg HistogramMode,    
  output reg Tri,
  output reg [2:0]  LB_STATE,       
  output reg [`REG_WIDTH-1:0] Dout,    
  output     [`REG_WIDTH-1:0] Din,
  output reg [10:0] ethLength,
  output reg ethLenOver,
  //
  // ethernet transmission controls
  input [2:0] packet_next,
  output reg ethTransmit, ethPacketReset,
  input ethPacketSent,
  output reg [4:0] ethHistSel,
  input [11:0] ethCounter

  );

  parameter 
  //  Parameters used for the STATEs of the LocBus FSM
  IDLE           =  0, // Waiting for the negative transition on LocCS_b
  RD_OR_WR       =  1, // check LocWRITE_b.  
  MASTER_WRITE   =  2, // If WR mode, then Store data
  MASTER_READ    =  3, // If RD mode, prepare data to be read
  END_OF_READ    =  4, // Latch data, release control lines, wait posedge of LocCS_b
  END_OF_WRITE   =  5, // Stop WR pulses and wait for posedge of LocCS_b
  END_CYCLE      =  6; // Set Data lines to HiZ (has to be after posedge)

//     INTERNAL SIGNALS
  reg  [`REG_WIDTH-1:0] CntrReg, StatusReg,WriteOnlyCntrReg, WriteOnlyCntrRegD;

  reg  LocCS_Del_0, LocCS_Del_1, LocCS_Del_2; 
  
//  ACTUAL CODE STARTS HERE
  assign  GateTTCcommands    =  CntrReg[11];

// forcing IOBUFfer to convice XST.  LocData goes stright to the pads while the rest of
//  the logic makes use of Din and Dout. NB: T=1 means Tristate.
//    _S_2  means Slow output, 2 mA drive (minimum, in case of bus collisions)
 IOBUF_S_2 LocDataBUF0  (.IO(LocData[0]),  .I(Dout[0]),  .O(Din[0]),  .T(Tri));
 IOBUF_S_2 LocDataBUF1  (.IO(LocData[1]),  .I(Dout[1]),  .O(Din[1]),  .T(Tri));
 IOBUF_S_2 LocDataBUF2  (.IO(LocData[2]),  .I(Dout[2]),  .O(Din[2]),  .T(Tri));
 IOBUF_S_2 LocDataBUF3  (.IO(LocData[3]),  .I(Dout[3]),  .O(Din[3]),  .T(Tri));
 IOBUF_S_2 LocDataBUF4  (.IO(LocData[4]),  .I(Dout[4]),  .O(Din[4]),  .T(Tri));
 IOBUF_S_2 LocDataBUF5  (.IO(LocData[5]),  .I(Dout[5]),  .O(Din[5]),  .T(Tri));
 IOBUF_S_2 LocDataBUF6  (.IO(LocData[6]),  .I(Dout[6]),  .O(Din[6]),  .T(Tri));
 IOBUF_S_2 LocDataBUF7  (.IO(LocData[7]),  .I(Dout[7]),  .O(Din[7]),  .T(Tri));
 IOBUF_S_2 LocDataBUF8  (.IO(LocData[8]),  .I(Dout[8]),  .O(Din[8]),  .T(Tri));
 IOBUF_S_2 LocDataBUF9  (.IO(LocData[9]),  .I(Dout[9]),  .O(Din[9]),  .T(Tri));
 IOBUF_S_2 LocDataBUF10 (.IO(LocData[10]), .I(Dout[10]), .O(Din[10]), .T(Tri));
 IOBUF_S_2 LocDataBUF11 (.IO(LocData[11]), .I(Dout[11]), .O(Din[11]), .T(Tri));
 IOBUF_S_2 LocDataBUF12 (.IO(LocData[12]), .I(Dout[12]), .O(Din[12]), .T(Tri));
 IOBUF_S_2 LocDataBUF13 (.IO(LocData[13]), .I(Dout[13]), .O(Din[13]), .T(Tri));
 IOBUF_S_2 LocDataBUF14 (.IO(LocData[14]), .I(Dout[14]), .O(Din[14]), .T(Tri));
 IOBUF_S_2 LocDataBUF15 (.IO(LocData[15]), .I(Dout[15]), .O(Din[15]), .T(Tri));
//
reg[2:0] ethHist;
always @ (ethHist or ethHistSel)
	case (ethHist)
		0: ethHistSel = 'h1;
		1: ethHistSel = 'h2;
		2: ethHistSel = 'h4;
		3: ethHistSel = 'h8;
		4: ethHistSel = 'h10;
	endcase
//
// delay VME_Hard_rst so that it doesn't freeze this module
//
reg VME_Hard_rst_1;
reg h1, h2, h3, h4, h5, h6, h7, h8;
always @ (posedge clk) begin
     h1 <= VME_Hard_rst_1;
	  h2 <= h1;
	  h3 <= h2;
	  h4 <= h4;
	  h5 <= h5;
	  h6 <= h6;
	  h7 <= h7;
	  h8 <= h8;
	  VME_Hard_rst <= h8;
   end
 //
 // this little trick allows you to read the fifo word correspdonging to the
 // fifo read cycle (as opposed to the way we've been doing it, which is you
 // read the fifo word on the next vme cycle AFTER the fifo read has been
 // issued).
 //
 reg read_fifo1, read_fifo2, read_fifo3;
 reg[35:0] fifo1, fifo2, fifo3;
 always @ (posedge clk) begin
 	read_fifo1 <= #1 TPSpyFifoRd[1];
 	read_fifo2 <= #1 TPSpyFifoRd[2];
 	read_fifo3 <= #1 TPSpyFifoRd[3];
   if (read_fifo1) fifo1 <= #1 TPSpyFIFO1_out;
   if (read_fifo2) fifo2 <= #1 TPSpyFIFO2_out;
   if (read_fifo3) fifo3 <= #1 TPSpyFIFO3_out;
	end
 always@(posedge clk)  begin  
   if(rst==1)   begin       //   RESET
 	  mux_out <= 0;
	  hist_localbus_read <= 0;
   end  //   END of RESET 
   else  begin   // posedge clk       
// WriteOnlyCntrReg self-reset itself and creates positive pulses (2 80-MHz ticks)
     WriteOnlyCntrRegD   <= WriteOnlyCntrReg;
     VME_Hard_rst_1    <= WriteOnlyCntrReg[0] | WriteOnlyCntrRegD[0];
     VME_Start       <= WriteOnlyCntrReg[1] | WriteOnlyCntrRegD[1];
     VME_Stop        <= WriteOnlyCntrReg[2] | WriteOnlyCntrRegD[2];
     VME_Soft_rst    <= WriteOnlyCntrReg[3] | WriteOnlyCntrRegD[3];
     VME_LVDSout     <= WriteOnlyCntrReg[5] | WriteOnlyCntrRegD[5] | CntrReg[0];
	  ethTransmit     <= WriteOnlyCntrReg[8] | WriteOnlyCntrRegD[8];
	  ethPacketReset  <= WriteOnlyCntrReg[7] | WriteOnlyCntrRegD[7];
	  read_count <= WriteOnlyCntrReg[6] | WriteOnlyCntrRegD[6];
	  HistogramMode   		<= CntrReg[1];
	  HZero 						<= CntrReg[2];
	  LEDdoBC0 					<= CntrReg[3];
	  hist_localbus_select 	<= CntrReg[8:4];
	  ethLenOver            <= CntrReg[10];
     LocCS_Del_0  <=  LocCS_b;    // FF to avoid metastability
     LocCS_Del_1  <=  LocCS_Del_0;
     LocCS_Del_2  <=  LocCS_Del_1;  // total of 3Tck = 75 ns delay on CS
     StatusReg[0]                 <=  RxDLL_lock;
     StatusReg[1]                 <=  run_status;  
     StatusReg[2]                 <=  BunchEr;
     StatusReg[3]                 <=  TTC_DLL_lock;
	  StatusReg[4]                 <=  HistogramsDone;
	  StatusReg[5]                 <=  ethPacketSent;
     StatusReg[7:6]               <=  0;
     StatusReg[10:8]              <=  UnlockCnt;
	  StatusReg[`REG_WIDTH-1:11]   <=  0;  // unused portion of  reg
    
    case(LB_STATE)
        
    IDLE:  // check for a negative edge on LocCS_b (digitally filtered)
        begin
          Dout      <=  'h BAD;  //mark a bad LocBus read timing
          Tri       <=  1;    // Data lines normally Hi-Z
          if(LocCS_Del_2 == 1 && LocCS_Del_1==0)  LB_STATE  <=  RD_OR_WR;
        end
            
    RD_OR_WR: //  decide whether it is a READ or a WRITE (master's point of view)
        begin
          if (LocWRITE_b==0)   LB_STATE <= MASTER_WRITE; 
          else  begin 
             LB_STATE <= MASTER_READ;  //  Read operation (master's point of view)
             Tri      <= 0;            //  Set Data lines toward LocalBus
            end  //  else
        end          
      //  ***********************************************   
      //     ********   W R I T E  operation  ********    
    MASTER_WRITE:  begin
       casez(LocAddr)    // decode which register is addressed                           
        'h0:      CntrReg           <=  Din[`REG_WIDTH-1:0];
        'h61:     WriteOnlyCntrReg  <=  Din[`REG_WIDTH-1:0];
		  'h14:     mux_out <= Din[7:0];
		  'h44:		TTC_BC0_pattern <= Din[7:0];
		  'h50:     begin
		  				hist_localbus_bin <= Din[11:0];
						hist_localbus_read <= 1;
						end
		  'h62: DestMAC[15:0] <= Din[15:0];
		  'h63: DestMAC[31:16] <= Din[15:0];
		  'h64: DestMAC[47:32] <= Din[15:0];
		  'h65: HLXMAC[15:0] <= Din[15:0];
		  'h66: HLXMAC[31:16] <= Din[15:0];
		  'h67: HLXMAC[47:32] <= Din[15:0];
	     'h70:     ethHist <= Din[2:0];
		  'h73: ethLength = Din[10:0];
       default:       ;
               //no action  if there is no such register or is read-only 
       endcase    //    (LocAddr)
       LB_STATE  <=  END_OF_WRITE;          
    end        //  end of MASTER_WRITE case

       //  ****************************************
       //    ******* R E A D  operation ********
    MASTER_READ:      begin
       casez(LocAddr)  // can use all Addr: 0 to 7F                                  
          // FIFO section: generate Fifo_RD pulses,
         // the data will be available on the next cycle
     
        'h35:   TPSpyFifoRd[1]     <=  1'b1;
        'h38:   TPSpyFifoRd[2]     <=  1'b1;
        'h3B:   TPSpyFifoRd[3]     <=  1'b1;
        default:       ;
       endcase   //  (LocAddr)
       LB_STATE  <=  END_OF_READ;          
      end      //  MASTER_READ  case
            
    END_OF_READ:  begin                   
       // release all fifo read enables

       TPSpyFifoRd      <=  0;

       casez(LocAddr)  // here we actually latch the data outputs 
        'h0:  Dout <= CntrReg;
        'h1:  Dout <= 'hFFFFF;  //  this reg is for testing
        'h2:  Dout <= 'h0;      //  this reg is for testing
        'h3:  begin
              Dout[`REG_WIDTH-1:`REG_WIDTH-4] <=  'h7; //Dout is 16 bits.  top 4 will be 'b0111
              Dout[`REG_WIDTH-5:0]  <=  `FPGAversionN;
		     end
//        'hA:  Dout <= {InstantCntErr[8:1], InstantR1sErr[8:1]};
//        'hB:  Dout <= AccCntErr;
//        'hC:  Dout <= AccR1sErr;
//        'hD:  Dout <= DLL_status;
        'h10: Dout <= StatusReg;
		  'h14: Dout <= mux_out;
        'h18: Dout <= {8'b0, LastTTCcommand[7:0]}; // EvN_res not there
        'h20: Dout <= TTC_test0;   
        'h21: Dout <= TTC_test1;
        'h22: Dout <= TTC_test2;
        'h23: Dout <= TTC_test3;
        'h24: Dout <= TTC_test4;
        'h25: Dout <= TTC_test5;
        'h26: Dout <= TTC_test6;
        'h27: Dout <= TTC_test7;
        'h28: Dout <= TTC_test8;
        'h29: Dout <= TTC_test9;
        'h2A: Dout <= TTC_testA;
        'h2B: Dout <= TTC_testB;
        'h2C: Dout <= TTC_testC;
        'h2D: Dout <= TTC_testD;
        'h2E: Dout <= TTC_testE;
        'h2F: Dout <= TTC_testF;                           
		  'h30: Dout <= sum0_count;
		  'h31: Dout <= sum1_count;
		  'h32: Dout <= sum2_count;
		  'h33: Dout <= sum3_count;
   
        'h35: Dout <=  {3'b0, TPSpyFifoEmpty[1],  fifo1[11:0]};      
        'h36: Dout <=  {4'b0,                     fifo1[23:12]};      
        'h37: Dout <=  {4'b0,                     fifo1[35:24]};      
        'h38: Dout <=  {3'b0, TPSpyFifoEmpty[2],  fifo2[11:0]};      
        'h39: Dout <=  {4'b0,                     fifo2[23:12]};      
        'h3A: Dout <=  {4'b0,                     fifo2[35:24]};      
        'h3B: Dout <=  {3'b0, TPSpyFifoEmpty[3],  fifo3[11:0]};      
        'h3C: Dout <=  {4'b0,                     fifo3[23:12]};      
        'h3D: Dout <=  {4'b0,                     fifo3[35:24]};      

        'h3E: Dout <=  RxClkCnt; 
        'h3F: Dout <=  TxBC0Cnt;     
        'h40: Dout <=  RxBC0Cnt; 
        'h41: Dout <=  TClkCnt;
		  'h42: Dout <=  count_bucket;
		  'h43: Dout <=  latch_bucket;
		  'h44: Dout <=  TTC_BC0_pattern;
		  'h50: Dout <= {4'h0,hist_localbus_bin[11:0]};
		  'h51: begin
		  		Dout <= hist0_value;
				hist_localbus_read <= 0;
				end
		  'h52: begin
		  		Dout <= hist1_value;
				hist_localbus_read <= 0;
				end
		  'h53: begin
		  		Dout <= hist2_value;
				hist_localbus_read <= 0;
				end
		  'h54: begin
		  		Dout <= hist3_value;
				hist_localbus_read <= 0;
				end
		  'h55: begin
		  		Dout <= histsum_value;
				hist_localbus_read <= 0;
				end
		  'h62: Dout <= DestMAC[15:0];
		  'h63: Dout <= DestMAC[31:16];
		  'h64: Dout <= DestMAC[47:32];
		  'h65: Dout <= HLXMAC[15:0];
		  'h66: Dout <= HLXMAC[31:16];
		  'h67: Dout <= HLXMAC[47:32];
	     'h70: Dout <= ethHist;
		  'h71: Dout <= packet_next;
		  'h72: Dout <= ethCounter;
		  'h73: Dout <= ethLength;


        default:  Dout <=  'H DEAD; // if there is no such read-reg, answer: DEAD 
       endcase              
       if(LocCS_Del_1==1)    LB_STATE    <=  END_CYCLE;
    end   // END_OF_READ 
          
    END_OF_WRITE:
       begin    
          WriteOnlyCntrReg <=  0;   // self-reset to generate 1-tick pulses
          if(LocCS_Del_1==1)   LB_STATE    <=  END_CYCLE;
       end
          
    END_CYCLE:          
             LB_STATE      <= IDLE;
    
    default: LB_STATE    <=  IDLE;
    endcase    //  end the case related to the LB_STATE variable
   end      // else posedge clk
   end        // always
endmodule