/** Separate storage for headers and packet body data.  
    There will be the minimum amount of storage for body data, but a full set of headers up to 64 packets,
	 which should allow for ample control messaging (header only)

    64*32 bytes -->	2048 bytes (one block ram!)
	 32 bytes == 8 32-bit words.  
*/

module ethernet_packet_buffer(wb_clk,wb_addr,wb_di,wb_do,wb_cycle,wb_strobe,
                                 wb_we,wb_ack,packet_header_data,packet_sel,packet_addr,packet_read,
                                 packet_write,packet_clk,packet_we,
											packet_store_mapping, // packet_sel <-> packet_write...
											histogram_data, histogram_addr, histogram_read_strobe, histogram_read_ack,
											slave_state,global_enable);
    input wb_clk;
    input [31:0] wb_addr;
    input [31:0] wb_di;
    output [31:0] wb_do;
    input wb_cycle;
    input wb_strobe;
    input wb_we;
    output wb_ack;
    input [5:0] packet_sel;
	 input packet_header_data;
    input [8:0] packet_addr;
    output [15:0] packet_read;
    input [15:0] packet_write;
    input packet_clk;
    input packet_we;
	 input packet_store_mapping;
	 input global_enable;

	 output [2:0] slave_state;

	output [15:0] histogram_addr;
	output reg histogram_read_strobe;
	input histogram_read_ack;
	input [31:0] histogram_data;
	reg [5:0] histogram_remap[63:0];

//   	reg [31:0] eth_buf_do_reg;
 	wire [31:0] swapped_histos;
	wire [31:0] header_do;
	wire [15:0] packet_header_read;
	reg is_header_wb;
	wire is_header_wb_cond;
	reg real_header_we;
	wire packet_header_we;

always @(posedge packet_clk)
	if (packet_store_mapping) histogram_remap[packet_sel]<=packet_write[5:0];

assign packet_read=(packet_header_data)?(packet_header_read):(16'h0);
assign packet_header_we=packet_we && packet_header_data && (packet_addr[8:4]==0);
assign wb_do=(is_header_wb)?(header_do):(swapped_histos);

assign swapped_histos[7:0]=histogram_data[15:8];
assign swapped_histos[15:8]=histogram_data[7:0];
assign swapped_histos[23:16]=histogram_data[31:24];
assign swapped_histos[31:24]=histogram_data[23:16];

	reg [2:0] slave_state;

parameter SLAVE_IDLE     =3'h1;
parameter SLAVE_REQ      =3'h2;
parameter SLAVE_REPORT   =3'h3;
parameter SLAVE_LATCH    =3'h4;
parameter SLAVE_WAITIDLE =3'h5;

	wire slave_start_cond;
assign slave_start_cond=(slave_state==SLAVE_IDLE && (wb_strobe && wb_cycle) && global_enable);

always @(posedge wb_clk)
	case (slave_state) 
	SLAVE_IDLE: if (slave_start_cond) begin
			if (is_header_wb_cond) slave_state<=SLAVE_LATCH; 
			else slave_state<=SLAVE_REQ;
		end else slave_state<=SLAVE_IDLE; // stay here
	SLAVE_LATCH : slave_state<=SLAVE_REPORT;
	SLAVE_REPORT : slave_state<=SLAVE_WAITIDLE;
	SLAVE_WAITIDLE : if (!histogram_read_ack) slave_state<=SLAVE_IDLE;
			 else slave_state<=SLAVE_WAITIDLE;
	SLAVE_REQ : if (histogram_read_ack) slave_state<=SLAVE_REPORT;
		    else slave_state<=SLAVE_REQ;
	default : slave_state<=SLAVE_IDLE;
	endcase

assign is_header_wb_cond=(wb_addr[10:5]==0);
always @(posedge wb_clk) 
	if (slave_start_cond) is_header_wb<=is_header_wb_cond;

always @(posedge wb_clk) 
	if (slave_start_cond) histogram_read_strobe<=!is_header_wb_cond;
	else if (slave_state==SLAVE_IDLE || slave_state==SLAVE_REPORT || slave_state==SLAVE_WAITIDLE) histogram_read_strobe<=1'h0;
	else histogram_read_strobe<=histogram_read_strobe;

assign wb_ack=(slave_state==SLAVE_REPORT);

always @(posedge wb_clk) begin
	real_header_we<=is_header_wb && wb_we && wb_cycle && wb_strobe && (slave_state==SLAVE_LATCH);
end

	wire [8:0] header_addr;
	wire [9:0] packet_header_addr;

assign header_addr={wb_addr[16:11],wb_addr[4:2]};
assign packet_header_addr[3:1]=packet_addr[3:1];
assign packet_header_addr[0]=!packet_addr[0]; // swap 16-bit words...
assign packet_header_addr[9:4]=packet_sel;

assign histogram_addr[0]=wb_addr[1]; // always zero...
assign histogram_addr[8:1]=wb_addr[10:2]-9'h8; // drop the first eight 32-bit words
assign histogram_addr[14:9]=histogram_remap[wb_addr[16:11]];
assign histogram_addr[15]=1'h0; // always! 

RAMB16_S18_S36 headerBuf(
	.DOB(header_do[31:0]),
	.ADDRB(header_addr),
	.CLKB(wb_clk),
	.DIB(wb_di),
	.ENB(1'b1),
	.SSRB(1'b0),
	.WEB(real_header_we),
 	.DIPB(4'b0),
	.DOA(packet_header_read),
	.DIA(packet_write),
	.ADDRA(packet_header_addr),
	.CLKA(packet_clk),
   .ENA(1'b1),      // Port B RAM Enable Input
   .SSRA(1'b0),    // Port B Synchronous Set/Reset Input
   .WEA(packet_header_we),       // Port B Write Enable Input
	.DIPA(2'b0)	
);



endmodule
